from __future__ import annotations

import inspect
import torch
from copy import deepcopy

CONFIG_ITEMS_FROM_STR = {
    'torch.float': torch.float,
    'torch.float32': torch.float32,
    'torch.float64': torch.float64,
    'torch.int': torch.int,
    'torch.int32': torch.int32,
    'torch.int64': torch.int64
}


def parse_wb_config(metadata: dict, cls: type | None = None, extra_remove_keys: tuple[str] = ()) -> dict:
    """ Convert metadata downloaded from W&B to a format that torch.nn.Module can use """
    # Copy because we don't want to modify the input
    config = deepcopy(metadata)

    # Replace 'torch.float32' -> `torch.float32`
    for key, value in config.items():
        if isinstance(value, str) and (value in CONFIG_ITEMS_FROM_STR):
            config[key] = CONFIG_ITEMS_FROM_STR[value]

    # Convert tuple to list
    for key, value in config.items():
        if isinstance(value, list):
            config[key] = tuple(value)

    # Remove keys that are not needed for the `cls`
    if cls is None:
        cls_remove_keys = []
    else:
        cls_argument_names = set(inspect.signature(cls).parameters.keys())
        cls_remove_keys = [key for key in config if key not in cls_argument_names]

    # Remove the keys
    remove_keys = set(list(extra_remove_keys) + cls_remove_keys)
    for key in remove_keys:
        del config[key]
    return config


def get_optimizer_cls(config: dict, key: str = 'optimizer') -> type:
    """ Return the optimizer class from the str-valued optimizer name
    Example
    --------
        get_optimizer_cls({'optimizer': 'Adam'}) -> torch.optim.Adam
    """
    return getattr(torch.optim, config[key])


def get_loss_cls(config: dict, key: str = 'loss_fn') -> type:
    """ Return the loss function class from the str-valued loss function name
    Example
    -------
        get_optimizer_cls({'loss_fn': 'CrossEntropyLoss'}) -> torch.nn.CrossEntropyLoss
    """
    return getattr(torch.nn, config[key])



from __future__ import annotations
from .directory import get_wb_directory, get_artifacts_directory

INCLUDE_STUBS = [
    '.py',
    '.ipynb',
    'requirements.txt',
    'README.md',
]

EXCLUDE_STUBS = [
    '.git',
    '.vscode',
    '.idea',
    get_wb_directory(),
    get_artifacts_directory()
]


def include_files(path: str) -> bool:
    """ Decide which file to include when code is logged """
    # See https://docs.wandb.ai/ref/python/run#log_code
    for stub in EXCLUDE_STUBS:
        if stub in path:
            return False

    for stub in INCLUDE_STUBS:
        if path.endswith(stub):
            return True
    return False



from __future__ import annotations
import os

PYTORCH_FRAMEWORK = 'pytorch'
TENSORFLOW_FRAMEWORK = 'tensorflow'
DEFAULT_FRAMEWORK = PYTORCH_FRAMEWORK


def set_framework(framework: str = DEFAULT_FRAMEWORK):
    os.environ['WBUTILS_FRAMEWORK'] = framework


def get_framework():
    if 'WBUTILS_FRAMEWORK' in os.environ:
        return os.environ['WBUTILS_FRAMEWORK']
    else:
        return DEFAULT_FRAMEWORK



import pytest
from wbutils import (get_wb_directory, get_artifacts_directory, get_temp_directory,
                     set_wb_directory, set_artifacts_directory, set_temp_directory)


def test_get_wb_directory():
    assert isinstance(get_wb_directory(), str)


def test_get_artifacts_directory():
    assert isinstance(get_artifacts_directory(), str)


def test_get_temp_directory():
    assert isinstance(get_temp_directory(), str)


def test_set_wb_directory():
    new_wb_directory = './.wb'
    set_wb_directory(new_wb_directory)
    assert get_wb_directory() == new_wb_directory


def test_set_artifacts_directory():
    new_artifacts_directory = './.artifacts'
    set_artifacts_directory(new_artifacts_directory)
    assert get_artifacts_directory() == new_artifacts_directory


def test_set_temp_directory():
    new_temp_directory = './.tmp'
    set_temp_directory(new_temp_directory)
    assert get_temp_directory() == new_temp_directory



import pytest
from wbutils import set_framework, get_framework


def test_get_framework():
    assert isinstance(get_framework(), str)


def test_set_framework():
    framework = 'pytorch'
    set_framework(framework)
    assert get_framework() == framework





import os
import logging

from wbutils.version import __version__
from wbutils.framework import set_framework, get_framework, PYTORCH_FRAMEWORK, TENSORFLOW_FRAMEWORK, DEFAULT_FRAMEWORK
from wbutils.include import include_files, INCLUDE_STUBS, EXCLUDE_STUBS
from wbutils.directory import (set_wb_directory, get_wb_directory,
                               set_artifacts_directory, get_artifacts_directory,
                               set_temp_directory, get_temp_directory)
from wbutils.io import get_temp_path_to_file, read_json, write_json, read_txt, write_txt, EXTENSION_TO_IO
from wbutils.artifact import create_artifact, load_data_from_artifact
import wbutils.parsing as parsing
import wbutils.parsing.pytorch
import wbutils.keys


_ROOT = os.path.abspath(os.path.dirname(__file__))


def get_path(partial_path=''):
    return os.path.join(_ROOT, partial_path)



from __future__ import annotations
import os
import ujson
import tempfile
import torch
from torch import nn, Tensor
from torch.utils.data import Dataset, DataLoader

from .directory import get_temp_directory


def get_temp_path_to_file(filename: str):
    path_to_tempdir = tempfile.mktemp(dir=get_temp_directory())
    os.makedirs(path_to_tempdir, exist_ok=True)
    return os.path.join(path_to_tempdir, filename)


def read_json(path_to_file: str):
    with open(path_to_file, 'r') as f:
        data = ujson.load(f)
    return data


def write_json(payload, path_to_file: str, indent=2):
    with open(path_to_file, 'w') as f:
        ujson.dump(payload, f, indent=indent)


def read_txt(path_to_file: str) -> str:
    with open(path_to_file, 'r') as f:
        lines = f.readlines()
    text = ''.join(lines)
    return text


def write_txt(payload, path_to_file: str):
    with open(path_to_file, 'w') as f:
        f.writelines(payload)


EXTENSION_TO_IO = {
    'txt': (read_txt, write_txt),
    'json': (read_json, write_json),
    'pt': (torch.load, torch.save),
    'pth': (torch.load, torch.save)
}



from __future__ import annotations

import os
import wandb as wb

from .io import get_temp_path_to_file, EXTENSION_TO_IO


def create_artifact(artifact_name: str, artifact_type: str, description: str, metadata,
                    files: dict[str, object], extension_to_io: dict = EXTENSION_TO_IO) -> wb.Artifact:
    if 'files' in metadata:
        metadata['files_'] = metadata['files']
    metadata['files'] = list(files.keys())  # We need list because `odict_keys` cannot be serialized

    artifact = wb.Artifact(name=artifact_name, type=artifact_type, description=description, metadata=metadata)
    for filename, payload in files.items():
        extension = os.path.splitext(filename)[-1][1:]
        _, write_fn = extension_to_io[extension]
        path_to_file = get_temp_path_to_file(filename)
        write_fn(payload, path_to_file)
        artifact.add_file(path_to_file)
    return artifact


def load_data_from_artifact(artifact: wb.Artifact, path_to_artifact_folder: str | None,
                            extension_to_io: dict = EXTENSION_TO_IO) -> dict:
    if path_to_artifact_folder is None:
        path_to_artifact_folder = artifact.download()
    files = {}
    for filename in artifact.metadata['files']:
        extension = os.path.splitext(filename)[-1][1:]
        read_fn, _ = extension_to_io[extension]
        path_to_file = os.path.join(path_to_artifact_folder, filename)
        payload = read_fn(path_to_file)
        files[filename] = payload
    return files



from __future__ import annotations

TRAIN_SPLIT = 'train'
VAL_SPLIT = 'val'
TEST_SPLIT = 'test'
CHECK_SPLIT = 'check'
SPLITS = [TRAIN_SPLIT, VAL_SPLIT, TEST_SPLIT]
SPLITS3 = [TRAIN_SPLIT, VAL_SPLIT, TEST_SPLIT]
SPLITS4 = [TRAIN_SPLIT, VAL_SPLIT, TEST_SPLIT, CHECK_SPLIT]


DATASET_ARTIFACT_TYPE = 'dataset'
MODEL_ARTIFACT_TYPE = 'model'
WEIGHTS_FILENAME = 'weights.pth'

LOSS = 'loss'
ACCURACY_PCT = 'acc_pct'
ACCURACY_FRAC = 'acc_frac'
EPOCH = 'n_epochs'
EXAMPLE = 'n_examples'
OPTIMIZER = 'optimizer'
BATCH_SIZE = 'batch_size'
LR = 'lr'
DROPOUT_RATE = 'dropout_rate'



from __future__ import annotations
import os


def set_wb_directory(path_to_dir: str = './.wb'):
    """ Use this function if you don't want to use the default './wandb' directory
    https://docs.wandb.ai/guides/track/environment-variables
    """
    os.makedirs(path_to_dir, exist_ok=True)
    os.environ['WANDB_DIR'] = path_to_dir


def set_artifacts_directory(path_to_dir: str = './.artifacts'):
    """ Use this function if you don't want to use the default './artifacts' directory
    https://docs.wandb.ai/guides/track/environment-variables
    """
    os.makedirs(path_to_dir, exist_ok=True)
    os.environ['WANDB_ARTIFACT_DIR'] = path_to_dir


def set_temp_directory(path_to_dir: str = './.tmp'):
    """ Use this function if you don't want to use the default './artifacts' directory
    https://docs.wandb.ai/guides/track/environment-variables
    """
    os.makedirs(path_to_dir, exist_ok=True)
    os.environ['WBUTILS_TEMP_DIR'] = path_to_dir


def get_wb_directory():
    # https://docs.wandb.ai/guides/track/environment-variables
    if 'WANDB_DIR' in os.environ:
        return os.environ['WANDB_DIR']
    else:
        return './wandb'


def get_artifacts_directory():
    # https://docs.wandb.ai/guides/track/environment-variables
    if 'WANDB_ARTIFACT_DIR' in os.environ:
        return os.environ['WANDB_ARTIFACT_DIR']
    else:
        return './.artifacts'


def get_temp_directory():
    if 'WBUTILS_TEMP_DIR' in os.environ:
        return os.environ['WBUTILS_TEMP_DIR']
    else:
        return './.tmp'
